@@ -1,3 +1,9 @@
+//! PMC - Power Management Controller
+//!
+//! At the moment, this module has limited support to allow for basic
+//! operation. It is currently used with a 12MHz oscillator to reach
+//! a CPU clock of 300MHz and an MCK of 150 MHz.
+
 use crate::efc::Efc;
 use crate::efc::FlashWaitStates;
 use crate::target_device::PMC;
@@ -5,6 +11,7 @@ use crate::target_device::PMC;
 pub use crate::target_device::pmc::pmc_mckr::MDIV_A as MckDivider;
 pub use crate::target_device::pmc::pmc_mckr::PRES_A as MckPrescaler;
 
+/// A HAL representation of the PMC peripheral
 pub struct Pmc {
     periph: PMC,
     settings: Option<ClockSettings>,
@@ -12,21 +19,31 @@ pub struct Pmc {
 
 #[derive(Debug, PartialEq, defmt::Format, Clone)]
 pub enum PmcError {
+    /// The given peripheral does not support PMC clocking
     ClockingError(PeripheralIdentifier),
+
+    /// The given configuration could not be made. This could be due
+    /// to an invalid request, or an unsupported configuration.
     InvalidConfiguration,
+
+    /// The current code does not support the requested action
     UnimplementedError,
+
+    /// An unexpected, unrecoverable error has occurred. Please
+    /// file an issue.
     InternalError,
 }
 
 /// The selected "Main Clock Oscillator" source
 ///
 /// TODO/NOTE: At the moment, we only support the internal trimmed 12MHz
-/// oscillator. Some driver behavior is hardcoded on this assumption for
-/// the sake of simplicity.
+/// oscillator OR an external 12MHz oscillator. Some driver behavior is
+// hardcoded on this assumption for the sake of simplicity.
 ///
 /// This corresponds to CKGR_MOR.MOSCSEL
 pub enum MainClockOscillatorSource {
-    MainCk12MHz,
+    MainRCOscInternal12MHz,
+    MainCrystalOscExternal12MHz
 }
 
 /// The selected "Master Clock" source
@@ -68,6 +85,7 @@ pub struct ClockSettings {
 }
 
 impl ClockSettings {
+    /// Calculate the resulting master clock from the given settings request
     pub fn calc_master_clk_mhz(&self) -> Result<u8, PmcError> {
         // NOTE: This is based on Figure 31-1 - "General Clock Distribution Block Diagram"
 
@@ -76,7 +94,8 @@ impl ClockSettings {
             MasterClockSource::PllaClock => {
                 // The internal 12MHz osc is currently the (only) choice for driving the PLLACK
                 let main_ck: f32 = match self.main_clk_osc_src {
-                    MainClockOscillatorSource::MainCk12MHz => 12.0,
+                    MainClockOscillatorSource::MainRCOscInternal12MHz => 12.0,
+                    MainClockOscillatorSource::MainCrystalOscExternal12MHz => 12.0,
                 };
 
                 // These values disable the PLLA
@@ -84,7 +103,7 @@ impl ClockSettings {
                     return Err(PmcError::InvalidConfiguration);
                 }
 
-                let plla_ck = main_ck * (self.multiplier_a as f32);
+                let plla_ck = main_ck * ((self.multiplier_a + 1) as f32);
                 let plla_ck = plla_ck / (self.divider_a as f32);
 
                 plla_ck
@@ -155,10 +174,12 @@ impl Pmc {
         }
     }
 
+    /// Obtain the currently configured settings
     pub fn settings(&self) -> Option<&ClockSettings> {
         self.settings.as_ref()
     }
 
+    /// Enable PMC clocking for the given peripheral(s).
     pub fn enable_peripherals(&mut self, pids: &[PeripheralIdentifier]) -> Result<(), PmcError> {
         if pids.is_empty() {
             return Ok(());
@@ -220,6 +241,10 @@ impl Pmc {
         Ok(())
     }
 
+    /// Set the PMC clock configuration
+    ///
+    /// NOTE: At the moment, this is only intended to be called once, at startup.
+    /// This is not enforced by software.
     pub fn set_clocks(&mut self, efc: &mut Efc, cfg: ClockSettings) -> Result<(), PmcError> {
         // Calculate the master clock to determine the number of flash wait states.
         // This must be done BEFORE increasing the clock speed, in case the current number
@@ -227,7 +252,7 @@ impl Pmc {
         //
         // The flash controller (EEFC) is driven from the master clock (stated in section 31.2)
         let mck_new = cfg.calc_master_clk_mhz()?;
-        let fws = FlashWaitStates::from_pclk_mhz(mck_new)?;
+        let fws = FlashWaitStates::from_mck_mhz(mck_new)?;
 
         efc.periph.eefc_wpmr.modify(|_r, w| {
             w.wpkey().passwd();
@@ -240,16 +265,86 @@ impl Pmc {
 
         // Note: This follows Datasheet 31.17 "Recommendeded Programming Sequence"
         //
-        // Steps 1-5 skipped, using the internal osc
-
-        // The Main RC oscillator. Three output frequencies can be selected: 4/8/12 MHz. By default 12 MHz is
-        // selected. 8 MHz and 12 MHz are factory-trimmed. The Main RC Oscillator is the default choice, and
-        // requires no further configuration.
+        // # Step 1
         //
-        // TODO: This is the only supported variant. This code will fail if we add another option.
-        // You should implement the logic of steps 1-5 here if you are adding more MCO source
-        // options to the public interface!
-        let MainClockOscillatorSource::MainCk12MHz = cfg.main_clk_osc_src;
+        // If the Main crystal oscillator is not required, the PLL and divider can be directly configured (Step 6.)
+        // else this oscillator must be started (Step 2.).
+        match cfg.main_clk_osc_src {
+            MainClockOscillatorSource::MainRCOscInternal12MHz => {
+                // The Main RC oscillator. Three output frequencies can be selected: 4/8/12 MHz. By default 12 MHz is
+                // selected. 8 MHz and 12 MHz are factory-trimmed. The Main RC Oscillator is the default choice, and
+                // requires no further configuration.
+                //
+                // NOTE: Only 12MHz mode is currently supported.
+            },
+            MainClockOscillatorSource::MainCrystalOscExternal12MHz => {
+                // # Step 2
+                //
+                // Enable the Main crystal oscillator by setting CKGR_MOR.MOSCXTEN. The user can define a
+                // startup time. This can be done by configuring the appropriate value in CKGR_MOR.MOSCXTST.
+                // Once this register has been correctly configured, the user must wait for PMC_SR.MOSCXTS to be
+                // set. This can be done either by polling PMC_SR.MOSCXTS, or by waiting for the interrupt line to
+                // be raised if the associated interrupt source (MOSCXTS) has been enabled in PMC_IER
+                //
+                // Note: When the crystal oscillator bypass is disabled (MOSCXTBY = 0), the MOSCXTS flag
+                // must be read at ‘0’ in PMC_SR before enabling the crystal oscillator (MOSCXTEN = 1).
+                if self.periph.pmc_sr.read().moscxts().bit_is_set() {
+                    defmt::println!("Crystal Oscillator already stabilized?");
+                    return Err(PmcError::InternalError);
+                }
+                self.periph.ckgr_mor.modify(|_r, w| {
+                    unsafe {
+                        // TODO: Can we wait less than the max time for this?
+                        //
+                        // This might be (partially) the cause of defmt detaching when using the
+                        // external clock.
+                        w.moscxtst().bits(0xFF);
+                    }
+                    w.moscxtby().clear_bit();
+                    w.moscxten().set_bit();
+                    w.key().passwd();
+                    w
+                });
+                while self.periph.pmc_sr.read().moscxts().bit_is_clear() {}
+
+                // # Step 3
+                //
+                // Switch MAINCK to the Main crystal oscillator by setting CKGR_MOR.MOSCSEL.
+                self.periph.ckgr_mor.modify(|_r, w| {
+                    w.key().passwd();
+                    w.moscsel().set_bit();
+                    w
+                });
+
+                // # Step 4
+                //
+                // Wait for PMC_SR.MOSCSELS to be set to ensure the switch is complete.
+                while self.periph.pmc_sr.read().moscsels().bit_is_clear() {}
+
+                // # Step 5
+                //
+                // Check MAINCK frequency:
+                // This frequency can be measured via CKGR_MCFR.
+                //
+                // Read CKGR_MCFR until the MAINFRDY field is set, after which the user can read
+                // CKGR_MCFR.MAINF by performing an additional read. This provides the number of Main clock
+                // cycles that have been counted during a period of 16 SLCK cycles.
+                while self.periph.ckgr_mcfr.read().mainfrdy().bit_is_clear() {}
+                let mainf = self.periph.ckgr_mcfr.read().mainf().bits();
+
+                // If MAINF = 0, switch MAINCK to the Main RC Oscillator by clearing CKGR_MOR.MOSCSEL. If
+                // MAINF ≠ 0, proceed to Step 6.
+                if mainf == 0 {
+                    defmt::println!("Error locking clock!");
+                    self.periph.ckgr_mor.modify(|_r, w| {
+                        w.moscsel().clear_bit();
+                        w.key().passwd();
+                        w
+                    });
+                    return Err(PmcError::InternalError);
+                }
+            },
+        }
 
         // # Step 6
         //
@@ -357,10 +452,14 @@ impl Pmc {
         // information, see "Clock Switching Waveforms".
         //
         // MCK is MAINCK divided by 2.
+        self.settings = Some(cfg);
         Ok(())
     }
 }
 
+/// A peripheral identifier
+/// This identifier represents the given hardware peripheral known by the
+/// PMC subsystem
 #[allow(non_camel_case_types)]
 #[derive(Debug, PartialEq, Eq, Copy, Clone, defmt::Format)]
 #[repr(u32)]
@@ -518,6 +617,7 @@ pub enum PeripheralIdentifier {
 }
 
 impl PeripheralIdentifier {
+    /// Does this identifier support PMC clocking?
     pub fn supports_pmc_clocking(&self) -> Result<(), ()> {
         use PeripheralIdentifier::*;
         // These pids don't support PMC clocking
